# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2008-2011 W. Trevor King <wking@drexel.edu>
#
# This file is part of calibcant.
#
# calibcant is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# calibcant is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with calibcant.  If not, see
# <http://www.gnu.org/licenses/>.

"""Surface bump analysis (measures photodiode sensitivity).

Separate the more general `bump_analyze()` from the other `bump_*()`
functions in calibcant.

The relevant physical quantities are:

* `Vzp_out`  Output z-piezo voltage (what we generate)
* `Vzp`      Applied z-piezo voltage (after external amplification)
* `Zp`       The z-piezo position
* `Zcant`    The cantilever vertical deflection
* `Vphoto`   The photodiode vertical deflection voltage (what we measure)

Which are related by the parameters:

* `zp_gain`           Vzp_out / Vzp
* `zp_sensitivity`    Zp / Vzp
* `photo_sensitivity` Vphoto / Zcant

`photo_sensitivity` is measured by bumping the cantilever against the
surface, where `Zp = Zcant` (see `calibrate.bump_acquire()`).  The
measured slope `Vphoto/Vout` is converted to `photo_sensitivity` with
`bump_analyze()`.

>>> import os
>>> from pprint import pprint
>>> import tempfile
>>> import numpy
>>> from .config import HDF5_BumpConfig
>>> from pypiezo.config import pprint_HDF5, HDF5_ChannelConfig, HDF5_AxisConfig

>>> fd,filename = tempfile.mkstemp(suffix='.h5', prefix='calibcant-')
>>> os.close(fd)

>>> bump_config = HDF5_BumpConfig(filename=filename, group='/bump/config/')
>>> bump_config.save()
>>> z_channel_config = HDF5_ChannelConfig(filename=None)
>>> z_channel_config['maxdata'] = 200
>>> z_channel_config['conversion-coefficients'] = (0,1)
>>> z_channel_config['conversion-origin'] = 0
>>> z_axis_config = HDF5_AxisConfig(filename=None)
>>> deflection_channel_config = HDF5_ChannelConfig(filename=None)
>>> deflection_channel_config['maxdata'] = 200
>>> deflection_channel_config['conversion-coefficients'] = (0,1)
>>> deflection_channel_config['conversion-origin'] = 0

>>> raw_bump = {
...     'z': numpy.arange(100, dtype=numpy.uint16),
...     'deflection': numpy.arange(100, dtype=numpy.uint16),
...     }
>>> raw_bump['deflection'][:50] = 50
>>> processed_bump = bump_analyze(
...     raw_bump, bump_config, z_channel_config, z_axis_config,
...     deflection_channel_config)
>>> bump_plot(data=raw_bump)  # TODO: convert to V and m
>>> bump_save(filename=filename, group='/bump/', raw_bump=raw_bump,
...     z_channel_config=z_channel_config, z_axis_config=z_axis_config,
...     deflection_channel_config=deflection_channel_config,
...     processed_bump=processed_bump)

>>> pprint_HDF5(filename)  # doctest: +ELLIPSIS, +REPORT_UDIFF
/
  /bump
    /bump/config
      /bump/config/deflection
        /bump/config/deflection/channel
          <HDF5 dataset "channel": shape (), type "|S1">
            0
          <HDF5 dataset "conversion-coefficients": shape (), type "|S4">
            0, 1
          <HDF5 dataset "conversion-origin": shape (), type "|S1">
            0
          <HDF5 dataset "device": shape (), type "|S12">
            /dev/comedi0
          <HDF5 dataset "inverse-conversion-coefficients": shape (), type "|S1">
<BLANKLINE>
          <HDF5 dataset "inverse-conversion-origin": shape (), type "|S3">
            1.0
          <HDF5 dataset "maxdata": shape (), type "|S3">
            200
          <HDF5 dataset "range": shape (), type "|S1">
            1
          <HDF5 dataset "subdevice": shape (), type "|S2">
            -1
      <HDF5 dataset "far-steps": shape (), type "|S3">
        200
      <HDF5 dataset "initial-position": shape (), type "|S6">
        -5e-08
      <HDF5 dataset "model": shape (), type "|S9">
        quadratic
      <HDF5 dataset "push-depth": shape (), type "|S5">
        2e-07
      <HDF5 dataset "push-speed": shape (), type "|S5">
        1e-06
      <HDF5 dataset "samples": shape (), type "|S4">
        1024
      <HDF5 dataset "setpoint": shape (), type "|S3">
        2.0
      /bump/config/z
        /bump/config/z/axis
          <HDF5 dataset "gain": shape (), type "|S3">
            1.0
          <HDF5 dataset "maximum": shape (), type "|S4">
            None
          <HDF5 dataset "minimum": shape (), type "|S4">
            None
          <HDF5 dataset "sensitivity": shape (), type "|S3">
            1.0
        /bump/config/z/channel
          <HDF5 dataset "channel": shape (), type "|S1">
            0
          <HDF5 dataset "conversion-coefficients": shape (), type "|S4">
            0, 1
          <HDF5 dataset "conversion-origin": shape (), type "|S1">
            0
          <HDF5 dataset "device": shape (), type "|S12">
            /dev/comedi0
          <HDF5 dataset "inverse-conversion-coefficients": shape (), type "|S1">
<BLANKLINE>
          <HDF5 dataset "inverse-conversion-origin": shape (), type "|S3">
            1.0
          <HDF5 dataset "maxdata": shape (), type "|S3">
            200
          <HDF5 dataset "range": shape (), type "|S1">
            1
          <HDF5 dataset "subdevice": shape (), type "|S2">
            -1
    <HDF5 dataset "processed": shape (), type "<f8">
      1.00000002115
    /bump/raw
      <HDF5 dataset "deflection": shape (100,), type "<u2">
        [50 50 ... 50 51 52 ... 97 98 99]
      <HDF5 dataset "z": shape (100,), type "<u2">
        [ 0  1  2  3  ... 97 98 99]

>>> (raw_bump,bump_config,z_channel_config,z_axis_config,
...  deflection_channel_config,processed_bump) = bump_load(
...     filename=filename, group='/bump/')

>>> pprint(raw_bump)  # doctest: +ELLIPSIS
{'deflection': array([50, 50, ... 51, 52, 53, ..., 97, 98, 99], dtype=uint16),
 'z': array([ 0,  1,  2,  ..., 97, 98, 99], dtype=uint16)}
>>> processed_bump  # doctest: +ELLIPSIS
1.00...

>>> os.remove(filename)
"""

import h5py as _h5py
import numpy as _numpy
from scipy.optimize import leastsq as _leastsq

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
    import time as _time  # for timestamping lines on plots
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from pypiezo.base import convert_bits_to_volts as _convert_bits_to_volts
from pypiezo.base import convert_bits_to_meters as _convert_bits_to_meters
from pypiezo.config import HDF5_ChannelConfig as _HDF5_ChannelConfig
from pypiezo.config import HDF5_AxisConfig as _HDF5_AxisConfig
from pypiezo.config import h5_create_group as _h5_create_group

from . import LOG as _LOG
from . import base_config as _base_config
from .config import Linear as _Linear
from .config import Quadratic as _Quadratic
from .config import HDF5_BumpConfig as _HDF5_BumpConfig


def bump_analyze(data, bump_config, z_channel_config, z_axis_config,
                 deflection_channel_config, plot=False):
    """Return the slope of the bump.

    Inputs:
      data              dictionary of data in DAC/ADC bits
      bump_config       `.config._BumpConfig` instance
      z_channel_config  z `pypiezo.config._ChannelConfig` instance
      z_axis_config     z `pypiezo.config._AxisConfig` instance
      deflection_channel_config
                        deflection `pypiezo.config._ChannelConfig` instance
      plot              boolean overriding matplotlib config setting.
    Returns:
      photo_sensitivity (Vphoto/Zcant) in Volts/m

    Checks for strong correlation (r-value) and low randomness chance
    (p-value).
    """
    z = _convert_bits_to_meters(z_channel_config, z_axis_config, data['z'])
    deflection = _convert_bits_to_volts(
        deflection_channel_config, data['deflection'])
    if bump_config['model'] == _Linear:
        kwargs = {
            'param_guesser': limited_linear_param_guess,
            'model': limited_linear,
            'sensitivity_from_fit_params': limited_linear_sensitivity,
            }
    else:  # _Quadratic
        kwargs = {
            'param_guesser': limited_quadratic_param_guess,
            'model': limited_quadratic,
            'sensitivity_from_fit_params': limited_quadratic_sensitivity,
            }
    photo_sensitivity = bump_fit(z, deflection, plot=plot, **kwargs)
    return photo_sensitivity

def limited_linear(x, params):
    """
    Model the bump as:
      flat region (off-surface)
      linear region (in-contact)
      flat region (high-voltage-rail)
    Parameters:
      x_contact (x value for the surface-contact kink)
      y_contact (y value for the surface-contact kink)
      slope (dy/dx at the surface-contact kink)
    """
    high_voltage_rail = 2**16 - 1 # bits
    x_contact,y_contact,slope = params
    y = slope*(x-x_contact) + y_contact
    y = _numpy.clip(y, y_contact, high_voltage_rail)
    return y

def limited_linear_param_guess(x, y):
    """
    Guess rough parameters for a limited_linear model.  Assumes the
    bump approaches (raising the deflection as it does so) first.
    Retracting after the approach is optional.  Approximates the contact
    position and an on-surface (high) position by finding first crossings
    of thresholds 0.3 and 0.7 of the y value's total range.  Not the
    most efficient algorithm, but it seems fairly robust.
    """
    y_contact = float(y.min())
    y_max = float(y.max())
    i = 0
    y_low  = y_contact + 0.3 * (y_max-y_contact)
    y_high = y_contact + 0.7 * (y_max-y_contact)
    while y[i] < y_low:
        i += 1
    i_low = i
    while y[i] < y_high:
        i += 1
    i_high = i
    x_contact = float(x[i_low])
    x_high = float(x[i_high])
    slope = (y_high - y_contact) / (x_high - x_contact)
    return (x_contact, y_contact, slope)

def limited_linear_sensitivity(params):
    """
    Return the estimated sensitivity to small deflections according to
    limited_linear fit parameters.
    """
    slope = params[2]
    return slope

def limited_quadratic(x, params):
    """
    Model the bump as:
      flat region (off-surface)
      quadratic region (in-contact)
      flat region (high-voltage-rail)
    Parameters:
      x_contact (x value for the surface-contact kink)
      y_contact (y value for the surface-contact kink)
      slope (dy/dx at the surface-contact kink)
      quad (d**2 y / dx**2, allow decreasing sensitivity with increased x)
    """
    high_voltage_rail = 2**16 - 1 # bits
    x_contact,y_contact,slope,quad = params
    y = slope*(x-x_contact) + quad*(x-x_contact)**2+ y_contact
    y = _numpy.clip(y, y_contact, high_voltage_rail)
    return y

def limited_quadratic_param_guess(x, y):
    """
    Guess rough parameters for a limited_quadratic model.  Assumes the
    bump approaches (raising the deflection as it does so) first.
    Retracting after the approach is optional.  Approximates the contact
    position and an on-surface (high) position by finding first crossings
    of thresholds 0.3 and 0.7 of the y value's total range.  Not the
    most efficient algorithm, but it seems fairly robust.
    """
    x_contact,y_contact,linear_slope = limited_linear_param_guess(x,y)
    contact_depth = x.max() - x_contact
    slope = linear_slope / 2
    quad = slope / contact_depth
    # The above slope and quad were chosen so
    #   x = contact_depth
    #   x*slope + x**2 * slope == x * linear_slope
    return (x_contact, y_contact, slope, quad)

def limited_quadratic_sensitivity(params):
    """
    Return the estimated sensitivity to small deflections according to
    limited_quadratic fit parameters.
    """
    slope = params[2]
    return slope

def bump_fit(z, deflection,
             param_guesser=limited_quadratic_param_guess,
             model=limited_quadratic,
             sensitivity_from_fit_params=limited_quadratic_sensitivity,
             plot=False):
    """Fit a aurface bump and return the photodiode sensitivitiy.

    Parameters:
      z              piezo position in meters
      deflection     photodiode deflection in volts
      param_guesser  function that guesses initial model parameters
      model          parametric model of deflection vs. z
      sensitivity_from_fit_params  given fit params, return the sensitivity
      plot              boolean overriding matplotlib config setting.
    Returns:
      photodiode_sensitivity  photodiode volts per piezo meter
    """
    _LOG.debug('fit bump data with model %s' % model)
    def residual(p, deflection, z):
        return model(z, p) - deflection
    param_guess = param_guesser(z, deflection)
    p,cov,info,mesg,ier = _leastsq(
        residual, param_guess, args=(deflection, z), full_output=True,
        maxfev=int(10e3))
    _LOG.debug('fitted params: %s' % p)
    _LOG.debug('covariance matrix: %s' % cov)
    #_LOG.debug('info: %s' % info)
    _LOG.debug('message: %s' % mesg)
    if ier == 1:
        _LOG.debug('solution converged')
    else:
        _LOG.debug('solution did not converge')
    if plot or _base_config['matplotlib']:
        yguess = model(z, param_guess)
        yfit = model(z, p)
        bump_plot({'z': z, 'deflection': deflection}, yguess=yguess, yfit=yfit)
    return sensitivity_from_fit_params(p)

def bump_save(filename, group='/', raw_bump=None, bump_config=None,
              z_channel_config=None, z_axis_config=None,
              deflection_channel_config=None, processed_bump=None):
    with _h5py.File(filename, 'a') as f:
        cwg = _h5_create_group(f, group)
        if raw_bump is not None:
            try:
                del cwg['raw/z']
            except KeyError:
                pass
            try:
                del cwg['raw/deflection']
            except KeyError:
                pass
            cwg['raw/z'] = raw_bump['z']
            cwg['raw/deflection'] = raw_bump['deflection']
        for config,key in [(bump_config, 'config/bump'),
                           (z_channel_config, 'config/z/channel'),
                           (z_axis_config, 'config/z/axis'),
                           (deflection_channel_config,
                            'config/deflection/channel')]:
            if config is None:
                continue
            config_cwg = _h5_create_group(cwg, key)
            config.save(group=config_cwg)
        if processed_bump is not None:
            try:
                del cwg['processed']
            except KeyError:
                pass
            cwg['processed'] = processed_bump

def bump_load(filename, group='/'):
    assert group.endswith('/')
    raw_bump = processed_bump = None
    configs = []
    with _h5py.File(filename, 'a') as f:
        try:
            raw_bump = {
                'z': f[group+'raw/z'][...],
                'deflection': f[group+'raw/deflection'][...],
                }
        except KeyError:
            pass
        for Config,key in [(_HDF5_BumpConfig, 'config/bump'),
                           (_HDF5_ChannelConfig, 'config/z/channel'),
                           (_HDF5_AxisConfig, 'config/z/axis'),
                           (_HDF5_ChannelConfig, 'config/deflection/channel')]:
            config = Config(filename=filename, group=group+key)
            configs.append(config)
        try:
            processed_bump = f[group+'processed'][...]
        except KeyError:
            pass
    ret = [raw_bump]
    ret.extend(configs)
    ret.append(processed_bump)
    for config in configs:
        config.load()
    return tuple(ret)

def bump_plot(data, yguess=None, yfit=None):
    "Plot the bump (Vphoto vs Vzp)"
    if not _matplotlib:
        raise _matplotlib_import_error
    figure = _matplotlib_pyplot.figure()
    if yfit is None:
        axes = figure.add_subplot(1, 1, 1)
    else:
        axes = figure.add_subplot(2, 1, 1)
        residual_axes = figure.add_subplot(2, 1, 2)
    timestamp = _time.strftime('%H%M%S')

    axes.hold(True)
    axes.plot(data['z'], data['deflection'], '.', label='bump')
    if yguess != None:
        axes.plot(data['z'], yguess, 'g-', label='guess')
    if yfit != None:
        axes.plot(data['z'], yfit, 'r-', label='fit')

    axes.set_title('bump surface %s' % timestamp)
    #axes.legend(loc='upper left')
    axes.set_xlabel('Z piezo (meters)')
    axes.set_ylabel('Photodiode (Volts)')
    if yfit is not None:
        # second subplot for residual
        residual_axes.plot(data['z'], data['deflection'] - yfit,
                           'r-', label='residual')
        #residual_axes.legend(loc='upper right')
        residual_axes.set_xlabel('Z piezo (meters)')
        residual_axes.set_ylabel('Photodiode (Volts)')
    figure.show()
